from datetime import datetime
# from symbol import parameters
from pydantic import BaseModel
from enum import Enum
# from backtesterrb30.libs.utils.historical_sources import HISTORICAL_SOURCES, HISTORICAL_INTERVALS_UNION
from backtesterrb30.libs.utils.timestamps import datetime_to_timestamp

class InstrumentFile(BaseModel):
    source: str
    instrument: str
    interval: str
    time_start: int
    time_stop: int

    @property
    def identifier(self):
        return self.source + "_" + self.instrument


    @classmethod
    def from_filename(cls, filename: str):
        if not filename or filename == '': 
            raise Exception('Empty file name name')
        if filename[-4:] != '.csv': 
            raise Exception('Bad file extension')
        filename = filename[:-4]
        source, instrument, interval, time_start, time_stop = tuple(filename.split('__'))
        if None in (source, instrument, interval, time_start, time_stop) or \
                '' in (source, instrument, interval, time_start, time_stop):
            raise Exception('Bad file name string provided')
        dict_instrument_file = {
            'source': source,
            'instrument': instrument,
            'interval': interval,
            'time_start': int(time_start),
            'time_stop': int(time_stop)
        }
        return cls(**dict_instrument_file)
    
    @classmethod
    def from_params(cls, 
                source: str, 
                instrument: str, 
                interval: Enum, 
                time_start: datetime,
                time_stop: datetime):
        dict_instrument_file = {
            'source': source,
            'instrument': instrument,
            'interval': interval.value,
            'time_start': datetime_to_timestamp(time_start),
            'time_stop': datetime_to_timestamp(time_stop)
        }
        return cls(**dict_instrument_file)

    def to_filename(self):
        return self.source + "__" \
            + self.instrument  + "__" \
            + self.interval  + "__" \
            + str(self.time_start)  + "__" \
            + str(self.time_stop) + ".csv"
        
    def __str__(self):
        return self.to_filename()

    def toJSON(self):
        return self.to_filename()

    # def get_file_name() -> str:



from backtesterrb30.libs.utils.run_service import run_service
microservice_name = 'python_backtester'
from backtesterrb30.python_backtester.python_backtester import Backtester
run_service(microservice_name, Backtester)


#!/usr/bin/env python3.7
# -*- coding: utf-8 -*-

from datetime import datetime, date
from decimal import Decimal
from enum import Enum
from typing import Dict, Iterable, List, Optional, Union, TypeVar
from urllib.parse import quote as urlencode

from backtesterrb30.libs.xnt.models.http_jto import dc, extract_to_model, timestamp_to_dt, str_to_dt
from backtesterrb30.libs.xnt.models.http_jto import Numeric, Serializable


class InstrumentType(Enum):
    STOCK = 'STOCK'
    FUTURE = 'FUTURE'
    BOND = 'BOND'
    FOREX = 'CURRENCY'
    FUND = 'FUND'
    OPTION = 'OPTION'
    CFD = 'CFD'
    CALENDAR_SPREAD = 'CALENDAR_SPREAD'
    FX_SPOT = 'FX_SPOT'


class Durations(Enum):
    day = 'day'
    good_till_cancel = 'good_till_cancel'
    good_till_time = 'good_till_time'
    immediate_or_cancel = 'immediate_or_cancel'
    fill_or_kill = 'fill_or_kill'
    at_the_opening = 'at_the_opening'
    at_the_close = 'at_the_close'
    unknown = 'unknown'


class OrderTypes(Enum):
    market = 'market'
    limit = 'limit'
    stop = 'stop'
    stop_limit = 'stop_limit'
    twap = 'twap'
    iceberg = 'iceberg'
    trailing_stop = 'trailing_stop'
    unknown = 'unknown'


class Side(Enum):
    BUY = 'buy'
    SELL = 'sell'


class PermissionStatus(Enum):
    blocked = "Blocked"
    full_access = 'Full'
    read_only = 'ReadOnly'
    close_only = 'CloseOnly'

    def __eq__(self, other) -> bool:
        return self.name == other.name

    def __ne__(self, other) -> bool:
        return self.name != other.name


class OptionRight(Enum):
    PULL = 'PUT'
    CALL = 'CALL'
    BOTH = "BOTH"


class CandleDurations(Enum):
    DAY1 = 86400
    HOUR1 = 3600
    MIN30 = 1800
    MIN5 = 300
    MIN1 = 60


class DataType(Enum):
    QUOTES = "quotes"
    TRADES = "trades"


class FeedLevel(Enum):
    BP = "best_price"
    MD = "market_depth"
    BEST_PRICE = "best_price"
    MARKET_DEPTH = "market_depth"


class Ordering(Enum):
    ASC = "ASC"
    DESC = "DESC"


class OrderStatuses(Enum):
    created = 'created'
    accepted = 'accepted'
    pending = 'pending'
    placing = 'placing'
    working = 'working'
    cancelled = 'cancelled'
    filled = 'filled'
    rejected = 'rejected'

    @staticmethod
    def terminated(other: Enum):
        return other in (OrderStatuses.filled, OrderStatuses.rejected, OrderStatuses.cancelled)

    @staticmethod
    def active(other: Enum):
        return not OrderStatuses.terminated(other)


class ModifyAction(Enum):
    REPLACE = 'replace'
    CANCEL = 'cancel'


class Scopes(Enum):
    symbols = "symbols"
    feed = "feed"
    change = "change"
    ohlc = "ohlc"
    crossrates = "crossrates"
    summary = "summary"
    orders = "orders"
    transactions = "transactions"
    accounts = "accounts"


class AuthMethods(Enum):
    JWT = "jwt"
    BASIC = "basic"


class QuoteV1(Serializable):

    def __init__(self, symbol_id: str, timestamp: int, bid: Numeric, ask: Numeric) -> None:
        self.timestamp = timestamp_to_dt(timestamp)
        self.symbol_id = symbol_id
        self.bid = dc(bid)
        self.ask = dc(ask)

    @property
    def mid(self) -> Optional[Decimal]:
        if (self.bid is None) or (self.ask is None):
            return None
        else:
            return (self.bid + self.ask) / dc(2)


class QuoteV2(Serializable):
    class Level(Serializable):
        def __init__(self, value: str, size: str) -> None:
            self.value = dc(value)
            self.size = dc(size)

        @property
        def price(self) -> Decimal:
            return self.value

    def __init__(self, symbol_id: str, timestamp: int, bid: List[Dict[str, Numeric]],
                 ask: List[Dict[str, Numeric]]) -> None:
        self.timestamp = timestamp_to_dt(timestamp)
        self.symbol_id = symbol_id
        self.bid = extract_to_model(bid, self.Level)
        self.ask = extract_to_model(ask, self.Level)

    @property
    def mid(self) -> Optional[Decimal]:
        if (self.bid is None) or (self.ask is None):
            return None
        else:
            return (self.bid[0].price + self.ask[0].price) / dc(2)


class QuoteV3(Serializable):
    class Level(Serializable):
        def __init__(self, price: str, size: str) -> None:
            self.price = dc(price)
            self.size = dc(size)

    def __init__(self, symbol_id: str, timestamp: int, bid: List[Dict[str, Numeric]],
                 ask: List[Dict[str, Numeric]]) -> None:
        self.timestamp = timestamp_to_dt(timestamp)
        self.symbol_id = symbol_id
        self.bid = extract_to_model(bid, self.Level)
        self.ask = extract_to_model(ask, self.Level)

    @property
    def mid(self) -> Optional[Decimal]:
        if (self.bid is None) or (self.ask is None):
            return None
        else:
            return (self.bid[0].price + self.ask[0].price) / dc(2)


QuoteType = TypeVar('QuoteType', QuoteV1, QuoteV2, QuoteV3, covariant=True)


class TradeV1(Serializable):
    def __init__(self, timestamp: int, symbol_id: str, value: str, size: str) -> None:
        self.timestamp = timestamp_to_dt(timestamp)
        self.symbol_id = symbol_id
        self.value = dc(value)
        self.size = dc(size)


class TradeV2(TradeV1):
    pass


class TradeV3(Serializable):
    def __init__(self, timestamp: int, symbol_id: str, price: str, size: str) -> None:
        self.timestamp = timestamp_to_dt(timestamp)
        self.symbol_id = symbol_id
        self.price = dc(price)
        self.size = dc(size)


TradeType = TypeVar('TradeType', TradeV1, TradeV2, TradeV3, covariant=True)


class ChangeV1(Serializable):
    def __init__(self, base_price: str, daily_change: str, symbol_id: str) -> None:
        self.base_price = dc(base_price)
        self.daily_change = dc(daily_change)
        self.symbol_id = symbol_id


class ChangeV2(ChangeV1):
    pass


class ChangeV3(Serializable):
    def __init__(self, last_session_close_price: str, daily_change: str, symbol_id: str) -> None:
        self.last_session_close_price = dc(last_session_close_price)
        self.daily_change = dc(daily_change)
        self.symbol_id = symbol_id


ChangeType = TypeVar('ChangeType', ChangeV1, ChangeV2, ChangeV3, covariant=True)


class OHLCQuotes(Serializable):
    def __init__(self, open_: Numeric, low: Numeric, high: Numeric, close: Numeric, timestamp: int) -> None:
        self.open_ = dc(open_)
        self.low = dc(low)
        self.high = dc(high)
        self.close = dc(close)
        self.timestamp = timestamp_to_dt(timestamp)


class OHLCTrades(OHLCQuotes):
    def __init__(self, open_: Numeric, low: Numeric, high: Numeric, close: Numeric, timestamp: int,
                 volume: Numeric) -> None:
        super().__init__(open_, low, high, close, timestamp)
        self.volume = dc(volume)


class Crossrate(Serializable):
    def __init__(self, pair: str, symbol_id: str, rate: str) -> None:
        self.pair = pair
        self.symbol_id = symbol_id
        self.rate = rate


class SummaryV1(Serializable):
    class CurrencyPos(Serializable):
        def __init__(self, code: str, value: float, converted_value: float) -> None:
            self.code = code
            self.value = dc(value)
            self.converted_value = dc(converted_value)

    class Position(Serializable):
        def __init__(self, id_: str, symbol_type: str, currency: str, price: float, average_price: float,
                     quantity: float, value: float, converted_value: float, pnl: float, converted_pnl: float) -> None:
            self.id_ = id_
            self.symbol_type = Serializable.to_enum(symbol_type, InstrumentType)
            self.currency = currency
            self.price = dc(price)
            self.average_price = dc(average_price)
            self.quantity = dc(quantity)
            self.value = dc(value)
            self.converted_value = dc(converted_value)
            self.pnl = dc(pnl)
            self.converted_pnl = dc(converted_pnl)

    def __init__(self, account: str, timestamp: int, currency: str, margin_utilization: float, free_money: float,
                 net_asset_value: float, money_used_for_margin: float, session_date: str, currencies: List[str],
                 positions: List[Dict[str, Union[str, float]]]) -> None:
        self.account = account
        self.positions = extract_to_model(positions, self.Position)
        self.currencies = extract_to_model(currencies, self.CurrencyPos)
        self.session_date = session_date
        self.free_money = dc(free_money)
        self.timestamp = timestamp_to_dt(timestamp)
        self.net_asset_value = dc(net_asset_value)
        self.margin_utilization = dc(margin_utilization)
        self.money_used_for_margin = dc(money_used_for_margin)
        self.currency = currency


class SummaryV2(Serializable):
    class CurrencyPos(Serializable):
        def __init__(self, code: str, value: str, converted_value: str) -> None:
            self.code = code
            self.value = dc(value)
            self.converted_value = dc(converted_value)

    class Position(Serializable):
        def __init__(self, id_: str, symbol_type: str, currency: str, price: str, average_price: str, quantity: str,
                     value: str, converted_value: str, pnl: str, converted_pnl: str) -> None:
            self.id_ = id_
            self.symbol_type = Serializable.to_enum(symbol_type, InstrumentType)
            self.currency = currency
            self.price = dc(price)
            self.average_price = dc(average_price)
            self.quantity = dc(quantity)
            self.value = dc(value)
            self.converted_value = dc(converted_value)
            self.pnl = dc(pnl)
            self.converted_pnl = dc(converted_pnl)

    def __init__(self, account: str, timestamp: int, currency: str, margin_utilization: str, free_money: str,
                 net_asset_value: str, money_used_for_margin: str, session_date: Optional[List[int]],
                 currencies: List[Dict[str, str]],  positions: List[Dict[str, str]]) -> None:
        self.account = account
        self.positions = extract_to_model(positions, self.Position)
        self.currencies = extract_to_model(currencies, self.CurrencyPos)
        self.session_date = datetime(*session_date) if session_date else None
        self.free_money = dc(free_money)
        self.timestamp = timestamp_to_dt(timestamp)
        self.net_asset_value = dc(net_asset_value)
        self.margin_utilization = dc(margin_utilization)
        self.money_used_for_margin = dc(money_used_for_margin)
        self.currency = currency


class SummaryV3(Serializable):
    class CurrencyPos(Serializable):
        def __init__(self, code: str, price: str, converted_value: str) -> None:
            self.code = code
            self.price = dc(price)
            self.converted_value = dc(converted_value)

    class Position(Serializable):
        def __init__(self, symbol_id: str, symbol_type: str, currency: str, price: str,
                     average_price: str, quantity: str, value: str, converted_value: str,
                     pnl: str, converted_pnl: str) -> None:
            self.symbol_id = symbol_id
            self.symbol_type = Serializable.to_enum(symbol_type, InstrumentType)
            self.currency = currency
            self.price = dc(price)
            self.average_price = dc(average_price)
            self.quantity = dc(quantity)
            self.value = dc(value)
            self.converted_value = dc(converted_value)
            self.pnl = dc(pnl)
            self.converted_pnl = dc(converted_pnl)

    def __init__(self, account_id: str, timestamp: int, currency: str, margin_utilization: float, free_money: float,
                 net_asset_value: float, money_used_for_margin: float, session_date: str, currencies: List[str],
                 positions: List[Dict[str, str]]) -> None:
        self.account_id = account_id
        self.positions = extract_to_model(positions, self.Position)
        self.currencies = extract_to_model(currencies, self.CurrencyPos)
        self.session_date = session_date
        self.free_money = dc(free_money)
        self.timestamp = timestamp_to_dt(timestamp)
        self.net_asset_value = dc(net_asset_value)
        self.margin_utilization = dc(margin_utilization)
        self.money_used_for_margin = dc(money_used_for_margin)
        self.currency = currency


SummaryType = TypeVar('SummaryType', SummaryV1, SummaryV2, SummaryV3, covariant=True)


class TransactionV1(Serializable):
    def __init__(self, operation_type: str, id_: str, asset: Optional[str], when: int, sum_: float,
                 symbol_id: Optional[str] = None, account_id: Optional[str] = None,
                 order_id: Optional[str] = None, order_pos: Optional[int] = None,
                 uuid_: Optional[str] = None, value_date: Optional[str] = None) -> None:
        self.operation_type = operation_type
        self.id_ = id_
        self.asset = asset
        self.when = timestamp_to_dt(when)
        self.sum_ = sum_
        self.symbol_id = symbol_id
        self.account_id = account_id
        self.order_id = order_id
        self.order_pos = dc(order_pos)
        self.uuid = uuid_
        self.value_date = date.fromisoformat(value_date) if value_date else None  # type: date


class TransactionV2(TransactionV1):
    pass


class TransactionV3(Serializable):
    def __init__(self, operation_type: Optional[str] = None, id_: Optional[str] = None, timestamp: Optional[int] = None,
                 sum_: Optional[float] = None, asset: Optional[str] = None, symbol_id: Optional[str] = None,
                 account_id: Optional[str] = None, order_id: Optional[str] = None, order_pos: Optional[int] = None,
                 uuid_: Optional[str] = None, value_date: Optional[str] = None) -> None:
        self.operation_type = operation_type
        self.id_ = id_
        self.asset = asset
        self.timestamp = timestamp_to_dt(timestamp)
        self.sum_ = sum_
        self.symbol_id = symbol_id
        self.account_id = account_id
        self.order_id = order_id
        self.order_pos = dc(order_pos)
        self.uuid = uuid_
        self.value_date = date.fromisoformat(value_date) if value_date else None  # type: date


TransactionType = TypeVar('TransactionType', TransactionV1, TransactionV2, TransactionV3, covariant=True)


class UserAccount(Serializable):
    def __init__(self, status: Union[str, PermissionStatus], account_id: str) -> None:
        self.status = Serializable.to_enum(status, PermissionStatus)
        self.account_id = account_id


class Identifiers(Serializable):
    def __init__(self, isin: Optional[str] = None, figi: Optional[str] = None, cusip: Optional[str] = None,
                 ric: Optional[str] = None, sedol: Optional[str] = None) -> None:
        self.isin = isin
        self.figi = figi
        self.cusip = cusip
        self.ric = ric
        self.sedol = sedol


class SymbolV1(Serializable):
    class OptionData(Serializable):
        def __init__(self, option_group_id: str, right: str, strike_price: float) -> None:
            self.option_group_id = option_group_id
            self.right = Serializable.to_enum(right, OptionRight)
            self.strike_price = dc(strike_price)

    def __init__(self, name: str, description: str, country: str, exchange: str, id_: str,
                 currency: str, mpi: float, type_: str, ticker: str, group: Optional[str] = None,
                 option_data: Optional[Dict[str, str]] = None, expiration: Optional[int] = None) -> None:
        self.option_data = extract_to_model(option_data, self.OptionData)
        self.i18n = {}
        self.name = name
        self.description = description
        self.country = country
        self.exchange = exchange
        self.id_ = id_
        self.currency = currency
        self.mpi = dc(mpi)
        self.type_ = Serializable.to_enum(type_, InstrumentType)
        self.ticker = ticker
        self.expiration = timestamp_to_dt(expiration)
        self.group = group


class SymbolV2(Serializable):
    class OptionData(Serializable):
        def __init__(self, option_group_id: str, right: str, strike_price: str) -> None:
            self.option_group_id = option_group_id
            self.right = Serializable.to_enum(right, OptionRight)
            self.strike_price = dc(strike_price)

    def __init__(self, name: str, description: str, country: str, exchange: str, id_: str,
                 currency: str, mpi: str, type_: str, ticker: str, group: Optional[str] = None,
                 option_data: Optional[Dict[str, str]] = None, expiration: Optional[int] = None,
                 identifiers: Optional[Dict[str, str]] = None) -> None:
        self.option_data = extract_to_model(option_data, self.OptionData)
        self.i18n = {}
        self.name = name
        self.description = description
        self.country = country
        self.exchange = exchange
        self.id_ = id_
        self.currency = currency
        self.mpi = dc(mpi)
        self.type_ = Serializable.to_enum(type_, InstrumentType)
        self.ticker = ticker
        self.expiration = timestamp_to_dt(expiration)
        self.group = group
        self.identifiers = extract_to_model(identifiers, Identifiers)


class SymbolV3(Serializable):
    class OptionData(Serializable):
        def __init__(self, option_group_id: str, option_right: str, strike_price: str) -> None:
            self.option_group_id = option_group_id
            self.option_right = Serializable.to_enum(option_right, OptionRight)
            self.strike_price = dc(strike_price)

    def __init__(self, name: str, description: str, country: str, exchange: str, symbol_id: str, currency: str,
                 min_price_increment: str, symbol_type: str, ticker: str, group: str, expiration: Optional[int] = None,
                 option_data: Optional[Dict[str, str]] = None, underlying_symbol_id: Optional[str] = None,
                 identifiers: Optional[Dict[str, str]] = None, icon: Optional[str] = None) -> None:
        self.option_data = extract_to_model(option_data, self.OptionData)
        self.name = name
        self.description = description
        self.country = country
        self.exchange = exchange
        self.symbol_id = symbol_id
        self.currency = currency
        self.min_price_increment = dc(min_price_increment)
        self.symbol_type = Serializable.to_enum(symbol_type, InstrumentType)
        self.ticker = ticker
        self.expiration = timestamp_to_dt(expiration)
        self.group = group
        self.underlying_symbol_id = underlying_symbol_id
        self.identifiers = extract_to_model(identifiers, Identifiers)
        self.icon = icon


SymbolType = TypeVar('SymbolType', SymbolV1, SymbolV2, SymbolV3, covariant=True)


class SymbolSpecification(Serializable):
    def __init__(self, leverage: Numeric, contract_multiplier: Numeric, price_unit: Numeric, units: str,
                 lot_size: Numeric) -> None:
        self.leverage = dc(leverage)
        self.contract_multiplier = dc(contract_multiplier)
        self.price_unit = dc(price_unit)
        self.units = units
        self.lot_size = dc(lot_size)


class Exchange(Serializable):
    def __init__(self, id_: str, name: str, country: str) -> None:
        self.id_ = id_
        self.name = name
        self.country = country


class Group(Serializable):
    def __init__(self, group: str, name: str, types: List[str], exchange: str) -> None:
        self.group = group
        self.name = name
        self.types = [Serializable.to_enum(ty, InstrumentType) for ty in types]
        self.exchange = exchange


class Schedule(Serializable):
    class Interval(Serializable):
        class Period(Serializable):
            def __init__(self, start: int, end: int) -> None:
                self.start = timestamp_to_dt(start)
                self.end = timestamp_to_dt(end)

        def __init__(self, name: str, period: Dict[str, int], order_types: Optional[Dict[str, str]] = None) -> None:
            self.name = name
            self.period = extract_to_model(period, self.Period)
            self.order_types = order_types

        @property
        def start(self) -> datetime:
            return self.period.start

        @property
        def end(self) -> datetime:
            return self.period.end

    def __init__(self, intervals: list) -> None:
        self.intervals = extract_to_model(intervals, self.Interval)


class OrderSentV1(Serializable):
    order_type = None  # type: OrderTypes

    def __init__(self, account: str, instrument: str, side: Side, quantity: Numeric, duration: Durations,
                 client_tag: Optional[str] = None, oco_group: Optional[str] = None,
                 if_done_parent_id: Optional[str] = None) -> None:
        self.account = account
        self.instrument = instrument
        self.side = side
        self.quantity = dc(quantity)
        self.duration = duration
        self.client_tag = client_tag
        self.oco_group = oco_group
        self.if_done_parent_id = if_done_parent_id


class OrderSentV2(Serializable):
    order_type = None  # type: OrderTypes

    def __init__(self, account_id: str, instrument: str, side: Side, quantity: Numeric, duration: Durations,
                 client_tag: Optional[str] = None, oco_group: Optional[str] = None,
                 if_done_parent_id: Optional[str] = None, take_profit: Optional[Numeric] = None,
                 stop_loss: Optional[Numeric] = None) -> None:
        self.account_id = account_id
        self.instrument = instrument
        self.side = side
        self.quantity = dc(quantity)
        self.duration = duration
        self.client_tag = client_tag
        self.oco_group = oco_group
        self.if_done_parent_id = if_done_parent_id
        self.take_profit = take_profit
        self.stop_loss = stop_loss


class OrderSentV3(Serializable):
    order_type = None  # type: OrderTypes

    def __init__(self, account_id: str, symbol_id: str, side: Side, quantity: Numeric, duration: Durations,
                 client_tag: Optional[str] = None, oco_group: Optional[str] = None,
                 if_done_parent_id: Optional[str] = None, take_profit: Optional[Numeric] = None,
                 stop_loss: Optional[Numeric] = None) -> None:
        self.account_id = account_id
        self.symbol_id = symbol_id
        self.side = side
        self.quantity = dc(quantity)
        self.duration = duration
        self.client_tag = client_tag
        self.oco_group = oco_group
        self.if_done_parent_id = if_done_parent_id
        self.take_profit = take_profit
        self.stop_loss = stop_loss


OrderSentType = TypeVar('OrderSentType', OrderSentV1, OrderSentV2, OrderSentV3, covariant=True)


class OrderMarketV1(OrderSentV1):
    def __init__(self, account: str, instrument: str, side: Side, quantity: Numeric, duration: Durations,
                 client_tag: Optional[str] = None, oco_group: Optional[str] = None,
                 if_done_parent_id: Optional[str] = None) -> None:
        super().__init__(account, instrument, side, quantity, duration, client_tag, oco_group, if_done_parent_id)
        self.order_type = OrderTypes.market


class OrderMarketV2(OrderSentV2):
    def __init__(self, account_id: str, instrument: str, side: Side, quantity: Numeric, duration: Durations,
                 client_tag: Optional[str] = None, oco_group: Optional[str] = None,
                 if_done_parent_id: Optional[str] = None, take_profit: Optional[Numeric] = None,
                 stop_loss: Optional[Numeric] = None) -> None:
        super().__init__(account_id, instrument, side, quantity, duration, client_tag, oco_group, if_done_parent_id,
                         take_profit, stop_loss)
        self.order_type = OrderTypes.market


class OrderMarketV3(OrderSentV3):
    def __init__(self, account_id: str, instrument: str, side: Side, quantity: Numeric, duration: Durations,
                 client_tag: Optional[str] = None, oco_group: Optional[str] = None,
                 if_done_parent_id: Optional[str] = None, take_profit: Optional[Numeric] = None,
                 stop_loss: Optional[Numeric] = None) -> None:
        super().__init__(account_id, instrument, side, quantity, duration, client_tag, oco_group, if_done_parent_id,
                         take_profit, stop_loss)
        self.order_type = OrderTypes.market


OrderMarketType = TypeVar("OrderMarketType", OrderMarketV1, OrderMarketV2, OrderMarketV3, covariant=True)


class OrderLimitV1(OrderSentV1):
    def __init__(self, account: str, instrument: str, side: Side, quantity: Numeric, duration: Durations,
                 limit_price: Numeric, client_tag: Optional[str] = None, oco_group: Optional[str] = None,
                 if_done_parent_id: Optional[str] = None) -> None:
        super().__init__(account, instrument, side, quantity, duration, client_tag, oco_group, if_done_parent_id)
        self.order_type = OrderTypes.limit
        self.limit_price = dc(limit_price)


class OrderLimitV2(OrderSentV2):
    def __init__(self, account_id: str, instrument: str, side: Side, quantity: Numeric, duration: Durations,
                 limit_price: Numeric, client_tag: Optional[str] = None, oco_group: Optional[str] = None,
                 if_done_parent_id: Optional[str] = None, take_profit: Optional[Numeric] = None,
                 stop_loss: Optional[Numeric] = None) -> None:
        super().__init__(account_id, instrument, side, quantity, duration, client_tag, oco_group, if_done_parent_id,
                         take_profit, stop_loss)
        self.order_type = OrderTypes.limit
        self.limit_price = dc(limit_price)


class OrderLimitV3(OrderSentV3):
    def __init__(self, account_id: str, symbol_id: str, side: Side, quantity: Numeric, duration: Durations,
                 limit_price: Numeric, client_tag: Optional[str] = None, oco_group: Optional[str] = None,
                 if_done_parent_id: Optional[str] = None, take_profit: Optional[Numeric] = None,
                 stop_loss: Optional[Numeric] = None) -> None:
        super().__init__(account_id, symbol_id, side, quantity, duration, client_tag, oco_group, if_done_parent_id,
                         take_profit, stop_loss)
        self.order_type = OrderTypes.limit
        self.limit_price = limit_price


OrderLimitType = TypeVar("OrderLimitType", OrderLimitV1, OrderLimitV2, OrderLimitV3, covariant=True)


class OrderStopV1(OrderSentV1):
    def __init__(self, account: str, instrument: str, side: Side, quantity: Numeric, duration: Durations,
                 stop_price: Numeric, client_tag: Optional[str] = None, oco_group: Optional[str] = None,
                 if_done_parent_id: Optional[str] = None) -> None:
        super().__init__(account, instrument, side, quantity, duration, client_tag, oco_group, if_done_parent_id)
        self.order_type = OrderTypes.stop
        self.stop_price = stop_price


class OrderStopV2(OrderSentV2):
    def __init__(self, account_id: str, instrument: str, side: Side, quantity: Numeric, duration: Durations,
                 stop_price: Numeric, client_tag: Optional[str] = None, oco_group: Optional[str] = None,
                 if_done_parent_id: Optional[str] = None, take_profit: Optional[Numeric] = None,
                 stop_loss: Optional[Numeric] = None) -> None:
        super().__init__(account_id, instrument, side, quantity, duration, client_tag, oco_group, if_done_parent_id,
                         take_profit, stop_loss)
        self.order_type = OrderTypes.stop
        self.stop_price = stop_price


class OrderStopV3(OrderSentV3):
    def __init__(self, account_id: str, symbol_id: str, side: Side, quantity: Numeric, duration: Durations,
                 stop_price: Numeric, client_tag: Optional[str] = None, oco_group: Optional[str] = None,
                 if_done_parent_id: Optional[str] = None, take_profit: Optional[Numeric] = None,
                 stop_loss: Optional[Numeric] = None) -> None:
        super().__init__(account_id, symbol_id, side, quantity, duration, client_tag, oco_group, if_done_parent_id,
                         take_profit, stop_loss)
        self.order_type = OrderTypes.stop
        self.stop_price = stop_price


OrderStopType = TypeVar("OrderStopType", OrderStopV1, OrderStopV2, OrderStopV3, covariant=True)


class OrderStopLimitV1(OrderSentV1):
    def __init__(self, account: str, instrument: str, side: Side, quantity: Numeric, duration: Durations,
                 limit_price: Numeric, stop_price: Numeric, client_tag: Optional[str] = None,
                 oco_group: Optional[str] = None, if_done_parent_id: Optional[str] = None) -> None:
        super().__init__(account, instrument, side, quantity, duration, client_tag, oco_group, if_done_parent_id)
        self.order_type = OrderTypes.stop_limit
        self.limit_price = limit_price
        self.stop_price = stop_price


class OrderStopLimitV2(OrderSentV2):
    def __init__(self, account_id: str, instrument: str, side: Side, quantity: Numeric, duration: Durations,
                 limit_price: Numeric, stop_price: Numeric, client_tag: Optional[str] = None,
                 oco_group: Optional[str] = None, if_done_parent_id: Optional[str] = None,
                 take_profit: Optional[Numeric] = None, stop_loss: Optional[Numeric] = None) -> None:
        super().__init__(account_id, instrument, side, quantity, duration, client_tag, oco_group, if_done_parent_id,
                         take_profit, stop_loss)
        self.order_type = OrderTypes.stop_limit
        self.limit_price = limit_price
        self.stop_price = stop_price


class OrderStopLimitV3(OrderSentV3):
    def __init__(self, account_id: str, symbol_id: str, side: Side, quantity: Numeric, duration: Durations,
                 limit_price: Numeric, stop_price: Numeric, client_tag: Optional[str] = None,
                 oco_group: Optional[str] = None, if_done_parent_id: Optional[str] = None,
                 take_profit: Optional[Numeric] = None, stop_loss: Optional[Numeric] = None) -> None:
        super().__init__(account_id, symbol_id, side, quantity, duration, client_tag, oco_group, if_done_parent_id,
                         take_profit, stop_loss)
        self.order_type = OrderTypes.stop_limit
        self.limit_price = limit_price
        self.stop_price = stop_price


OrderStopLimitType = TypeVar("OrderStopLimitType", OrderStopLimitV1, OrderStopLimitV2, OrderStopLimitV3, covariant=True)


class OrderTrailingStopV1(OrderSentV1):
    def __init__(self, account: str, instrument: str, side: Side, quantity: Numeric, duration: Durations,
                 price_distance: Numeric, client_tag: Optional[str] = None, oco_group: Optional[str] = None,
                 if_done_parent_id: Optional[str] = None) -> None:
        super().__init__(account, instrument, side, quantity, duration, client_tag, oco_group, if_done_parent_id)
        self.order_type = OrderTypes.trailing_stop
        self.price_distance = price_distance


class OrderTrailingStopV2(OrderSentV2):
    def __init__(self, account_id: str, instrument: str, side: Side, quantity: Numeric, duration: Durations,
                 price_distance: Numeric, client_tag: Optional[str] = None, oco_group: Optional[str] = None,
                 if_done_parent_id: Optional[str] = None, take_profit: Optional[Numeric] = None,
                 stop_loss: Optional[Numeric] = None) -> None:
        super().__init__(account_id, instrument, side, quantity, duration, client_tag, oco_group, if_done_parent_id,
                         take_profit, stop_loss)
        self.order_type = OrderTypes.trailing_stop
        self.price_distance = price_distance


class OrderTrailingStopV3(OrderSentV3):
    def __init__(self, account_id: str, symbol_id: str, side: Side, quantity: Numeric, duration: Durations,
                 price_distance: Numeric, client_tag: Optional[str] = None, oco_group: Optional[str] = None,
                 if_done_parent_id: Optional[str] = None, take_profit: Optional[Numeric] = None,
                 stop_loss: Optional[Numeric] = None) -> None:
        super().__init__(account_id, symbol_id, side, quantity, duration, client_tag, oco_group, if_done_parent_id,
                         take_profit, stop_loss)
        self.order_type = OrderTypes.trailing_stop
        self.price_distance = price_distance


OrderTrailingStopType = TypeVar("OrderTrailingStopType", OrderTrailingStopV1, OrderTrailingStopV2, OrderTrailingStopV3,
                                covariant=True)


class OrderTwapV1(OrderSentV1):
    def __init__(self, account: str, instrument: str, side: Side, quantity: Numeric, duration: Durations,
                 part_quantity: Numeric, place_interval: Numeric, client_tag: Optional[str] = None,
                 oco_group: Optional[str] = None, if_done_parent_id: Optional[str] = None) -> None:
        super().__init__(account, instrument, side, quantity, duration, client_tag, oco_group, if_done_parent_id)
        self.order_type = OrderTypes.twap
        self.part_quantity = part_quantity
        self.place_interval = place_interval


class OrderTwapV2(OrderSentV2):
    def __init__(self, account_id: str, instrument: str, side: Side, quantity: Numeric, duration: Durations,
                 part_quantity: Numeric, place_interval: Numeric, client_tag: Optional[str] = None,
                 oco_group: Optional[str] = None, if_done_parent_id: Optional[str] = None,
                 take_profit: Optional[Numeric] = None, stop_loss: Optional[Numeric] = None) -> None:
        super().__init__(account_id, instrument, side, quantity, duration, client_tag, oco_group, if_done_parent_id,
                         take_profit, stop_loss)
        self.order_type = OrderTypes.twap
        self.part_quantity = part_quantity
        self.place_interval = place_interval


class OrderTwapV3(OrderSentV3):
    def __init__(self, account_id: str, symbol_id: str, side: Side, quantity: Numeric, duration: Durations,
                 part_quantity: Numeric, place_interval: Numeric, client_tag: Optional[str] = None,
                 oco_group: Optional[str] = None, if_done_parent_id: Optional[str] = None,
                 take_profit: Optional[Numeric] = None, stop_loss: Optional[Numeric] = None) -> None:
        super().__init__(account_id, symbol_id, side, quantity, duration, client_tag, oco_group, if_done_parent_id,
                         take_profit, stop_loss)
        self.order_type = OrderTypes.twap
        self.part_quantity = part_quantity
        self.place_interval = place_interval


OrderTwapType = TypeVar("OrderTwapType", OrderTwapV1, OrderTwapV2, OrderTwapV3, covariant=True)


class OrderIcebergV1(OrderSentV1):
    def __init__(self, account: str, instrument: str, side: Side, quantity: Numeric, duration: Durations,
                 part_quantity: Numeric, limit_price: Numeric, client_tag: Optional[str] = None,
                 oco_group: Optional[str] = None, if_done_parent_id: Optional[str] = None) -> None:
        super().__init__(account, instrument, side, quantity, duration, client_tag, oco_group, if_done_parent_id)
        self.order_type = OrderTypes.iceberg
        self.limit_price = limit_price
        self.part_quantity = part_quantity


class OrderIcebergV2(OrderSentV2):
    def __init__(self, account_id: str, instrument: str, side: Side, quantity: Numeric, duration: Durations,
                 part_quantity: Numeric, limit_price: Numeric, client_tag: Optional[str] = None,
                 oco_group: Optional[str] = None, if_done_parent_id: Optional[str] = None,
                 take_profit: Optional[Numeric] = None, stop_loss: Optional[Numeric] = None) -> None:
        super().__init__(account_id, instrument, side, quantity, duration, client_tag, oco_group, if_done_parent_id,
                         take_profit, stop_loss)
        self.order_type = OrderTypes.iceberg
        self.limit_price = limit_price
        self.part_quantity = part_quantity


class OrderIcebergV3(OrderSentV3):
    def __init__(self, account_id: str, symbol_id: str, side: Side, quantity: Numeric, duration: Durations,
                 part_quantity: Numeric, limit_price: Numeric, client_tag: Optional[str] = None,
                 oco_group: Optional[str] = None, if_done_parent_id: Optional[str] = None,
                 take_profit: Optional[Numeric] = None, stop_loss: Optional[Numeric] = None) -> None:
        super().__init__(account_id, symbol_id, side, quantity, duration, client_tag, oco_group, if_done_parent_id,
                         take_profit, stop_loss)
        self.order_type = OrderTypes.iceberg
        self.limit_price = limit_price
        self.part_quantity = part_quantity


OrderIcebergType = TypeVar("OrderIcebergType", OrderIcebergV1, OrderIcebergV2, OrderIcebergV3, covariant=True)


class Fill(Serializable):
    def __init__(self, quantity: str, price: str, position: int, time: str) -> None:
        self.quantity = dc(quantity)
        self.price = dc(price)
        self.position = position
        self.time = str_to_dt(time)


class OrderState(Serializable):
    def __init__(self, last_update: str, status: str, fills: List[Dict[str, Union[str, int]]],
                 reason: Optional[str] = None):
        self.last_update = str_to_dt(last_update)
        self.status = Serializable.to_enum(status, OrderStatuses)
        self.fills = extract_to_model(fills, Fill)
        # TODO: rewrite it to normal Reject model
        self.reason = reason


class Reject(Serializable):
    def __init__(self, group: str, message: str):
        self.group = group
        self.message = message


class OrderV1(Serializable):
    class OrderParameters(Serializable):
        def __init__(self, side: str, duration: str, quantity: str, instrument: str, order_type: str,
                     oco_group: Optional[str] = None, if_done_parent_id: Optional[str] = None,
                     limit_price: Optional[str] = None, stop_price: Optional[str] = None,
                     price_distance: Optional[str] = None, part_quantity: Optional[str] = None,
                     place_interval: Optional[str] = None) -> None:
            self.side = Serializable.to_enum(side, Side)
            self.duration = Serializable.to_enum(duration, Durations)
            self.quantity = dc(quantity)
            self.instrument = instrument
            self.order_type = Serializable.to_enum(order_type, OrderTypes)
            self.oco_group = oco_group
            self.if_done_parent_id = if_done_parent_id
            self.limit_price = dc(limit_price)
            self.stop_price = dc(stop_price)
            self.price_distance = dc(price_distance)
            self.part_quantity = dc(part_quantity)
            self.place_interval = dc(place_interval)

    def __init__(self, place_time: str, order_state: Dict[str, Union[List, str, None]], id_: str,
                 order_parameters: Dict[str, Optional[str]], current_modification_id: str, exante_account: str,
                 client_tag: Optional[str] = None, username: Optional[str] = None):
        self.place_time = str_to_dt(place_time)
        self.order_state = extract_to_model(order_state, OrderState)
        self.id_ = id_
        self.client_tag = client_tag
        self.order_parameters = extract_to_model(order_parameters, self.OrderParameters)
        self.current_modification_id = current_modification_id
        self.exante_account = exante_account
        self.username = username


class OrderV2(Serializable):
    class OrderParameters(Serializable):
        def __init__(self, side: str, duration: str, quantity: str, instrument: str, order_type: str,
                     oco_group: Optional[str] = None, if_done_parent_id: Optional[str] = None,
                     limit_price: Optional[str] = None, stop_price: Optional[str] = None,
                     price_distance: Optional[str] = None, part_quantity: Optional[str] = None,
                     place_interval: Optional[str] = None) -> None:
            self.side = Serializable.to_enum(side, Side)
            self.duration = Serializable.to_enum(duration, Durations)
            self.quantity = dc(quantity)
            self.instrument = instrument
            self.order_type = Serializable.to_enum(order_type, OrderTypes)
            self.oco_group = oco_group
            self.if_done_parent_id = if_done_parent_id
            self.limit_price = dc(limit_price)
            self.stop_price = dc(stop_price)
            self.price_distance = dc(price_distance)
            self.part_quantity = dc(part_quantity)
            self.place_interval = dc(place_interval)

    def __init__(self, place_time: str, order_state: Dict, id_: str, order_parameters: Dict,
                 username: str, current_modification_id: str, account_id: str = None,
                 client_tag: Optional[str] = None) -> None:
        self.place_time = str_to_dt(place_time)
        self.order_state = extract_to_model(order_state, OrderState)
        self.id_ = id_
        self.client_tag = client_tag
        self.order_parameters = extract_to_model(order_parameters, self.OrderParameters)
        self.current_modification_id = current_modification_id
        self.account_id = account_id
        self.username = username


class OrderV3(Serializable):
    class OrderParameters(Serializable):
        def __init__(self, side: str, duration: str, quantity: str, symbol_id: str, order_type: str,
                     oco_group: Optional[str] = None, if_done_parent_id: Optional[str] = None,
                     limit_price: Optional[str] = None, stop_price: Optional[str] = None,
                     price_distance: Optional[str] = None, part_quantity: Optional[str] = None,
                     place_interval: Optional[str] = None) -> None:
            self.side = Serializable.to_enum(side, Side)
            self.duration = Serializable.to_enum(duration, Durations)
            self.quantity = dc(quantity)
            self.symbol_id = symbol_id
            self.order_type = Serializable.to_enum(order_type, OrderTypes)
            self.oco_group = oco_group
            self.if_done_parent_id = if_done_parent_id
            self.limit_price = dc(limit_price)
            self.stop_price = dc(stop_price)
            self.price_distance = dc(price_distance)
            self.part_quantity = dc(part_quantity)
            self.place_interval = dc(place_interval)

    def __init__(self, place_time: str, order_state: Dict, order_id: str, order_parameters: Dict, username: str,
                 current_modification_id: str, account_id: str, client_tag: Optional[str] = None) -> None:
        self.place_time = str_to_dt(place_time)
        self.order_state = extract_to_model(order_state, OrderState)
        self.order_id = order_id
        self.client_tag = client_tag
        self.order_parameters = extract_to_model(order_parameters, self.OrderParameters)
        self.current_modification_id = current_modification_id
        self.account_id = account_id
        self.username = username


OrderType = TypeVar('OrderType', OrderV1, OrderV2, OrderV3, covariant=True)


class ExOrderV1(Serializable):
    def __init__(self, quantity: str, order_id: str, event: str, price: str, position: str, time: str,) -> None:
        self.quantity = dc(quantity)
        self.order_id = order_id
        self.event = event
        self.price = dc(price)
        self.position = dc(position)
        self.time = str_to_dt(time)


class ExOrderV2(ExOrderV1):
    pass


class ExOrderV3(Serializable):
    def __init__(self, quantity: str, order_id: str, price: str, position: str, timestamp: str) -> None:
        self.quantity = dc(quantity)
        self.order_id = order_id
        self.price = dc(price)
        self.position = dc(position)
        self.timestamp = str_to_dt(timestamp)


ExOrderType = TypeVar("ExOrderType", ExOrderV1, ExOrderV2, ExOrderV3, covariant=True)


def resolve_model(version: str,
                  class_type: Union[TradeType, QuoteType, ChangeType, TradeType, SummaryType, OrderSentType,
                                    OrderMarketType, OrderLimitType, OrderStopType, OrderStopLimitType,
                                    OrderTrailingStopType, OrderTwapType, OrderIcebergType]):
    classes = list(filter(lambda x: f"{class_type.__name__[:-4]}V{version.split('.')[0]}" == x.__name__,
                          class_type.__constraints__))  # type: List[Serializable]
    if len(classes) > 1:
        raise Exception("Unexpected resolution!")
    else:
        return classes[0]


def resolve_symbol(symbol: Union[str, SymbolType, Iterable[str], Iterable[SymbolType]]) -> Optional[str]:
    if isinstance(symbol, SymbolV3):
        return urlencode(symbol.symbol_id, safe="")
    elif isinstance(symbol, (SymbolV1, SymbolV2)):
        return urlencode(symbol.id_, safe="")
    elif isinstance(symbol, Iterable) and not isinstance(symbol, str):
        return ",".join([resolve_symbol(x) for x in symbol])
    elif symbol:
        return urlencode(symbol, safe="")
    else:
        return symbol



from typing import Union
from backtesterrb30.libs.xnt.http_api import HTTPApi, current_api, AuthMethods, CandleDurations, DataType
import pandas as pd
from os.path import join
from time import time, sleep
from datetime import datetime
from backtesterrb30.libs.data_sources.data_source_base import DataSource
from backtesterrb30.libs.interfaces.historical_data_feeds.instrument_file import InstrumentFile
from backtesterrb30.libs.interfaces.utils.data_symbol import DataSymbol
# from backtesterrb30.libs.utils.historical_sources import EXANTE_INTERVALS
from backtesterrb30.libs.utils.singleton import singleton
from backtesterrb30.libs.data_sources.utils import validate_dataframe_timestamps
from os import getenv
import asyncio
from enum import Enum

class EXANTE_INTERVALS_2(str, Enum):
    tick: str='tick'
    minute: str='minute'
    minute5: str='minute5'
    minute30: str='minute30'
    hour: str='hour'
    day: str='day'

class ExanteDataSource(DataSource):
    INTERVALS= EXANTE_INTERVALS_2
    NAME='exante'
    
    def __init__(self, logger=print):
        super().__init__(True, logger)
        exante_app_id=getenv("exante_app_id")
        exante_access_key=getenv("exante_access_key")
        if exante_app_id == None or exante_access_key == None:
            raise Exception(self.NAME + ' No authorization heys provided')
        self.client = HTTPApi(AuthMethods.BASIC, appid=exante_app_id, acckey=exante_access_key)

    def __get_exante_interval(self, interval: str) -> CandleDurations:
        if interval == 'minute': return CandleDurations.MIN1
        if interval == 'minute5': return CandleDurations.MIN5
        if interval == 'minute30': return CandleDurations.MIN30
        if interval == 'hour': return CandleDurations.HOUR1
        if interval == 'day': return CandleDurations.DAY1

    def _get_interval_miliseconds(self, interval: str) -> Union[int,None]: 
        return int(self.__get_exante_interval(interval).value * 1000)

    #override
    async def _validate_instrument_data(self, data: DataSymbol) -> bool:
        #TODO check if volume necessery or not.

        data_type = DataType.QUOTES
        if data.with_volume == True:
            data_type = DataType.TRADES

        start_validation_time = time()
        candles = None
        while True:
            candles = self.client.get_ohlc(symbol=data.symbol, 
                        duration=60, start=1, version='3.0', 
                        limit=1, agg_type=data_type)
            if candles != None or time() - start_validation_time > 61:
                break
            self._log('Performing Exante validation for this instrument going to take up to 1 minute')
            await asyncio.sleep(10)
        if candles == None:
            self._log('Error. Instrument "'+data.symbol+'" probably does not exists on exante with this configuration. Try without volume')
            return False
        # from_datetime_timestamp = int(from_datetime.timestamp() * 1000)
        first_datetime = candles[0].timestamp
        if first_datetime > data.backtest_date_start:
            self._log("Error. First avaliable date of " , data.symbol, "is" , first_datetime)
            return False
        return True

    async def __wait(self):
        self._log('waiting for exante download 60s')
        await asyncio.sleep(60)

    #override
    async def _download_instrument_data(self, 
                    instrument: str, interval: str, time_start: int, time_stop: Union[int, None]) -> pd.DataFrame:
        self._log('Downloading exante data', instrument)
        # try:
        await self.__wait()
        if interval == 'tick': 
            limit = 5000
            floating_start = time_start
            ticks_arr = []
            iteration_n = 0
            while True:
                if floating_start >= time_stop: break
                ticks = self.client.get_ticks(symbol=instrument, 
                                    start = floating_start, 
                                    end = time_stop, 
                                    limit = limit, 
                                    agg_type=DataType.QUOTES)
                if ticks == None:
                    self._raise_error("Error while downloading ticks")
                ticks_arr = ticks + ticks_arr
                if ticks == []:
                    if iteration_n == 0:
                        self._raise_error("Error while downloading ticks ...")
                if len(ticks) == 5000:
                    floating_start = int(ticks[0].timestamp.timestamp() * 1000) + 1

                if len(ticks)<5000:
                    break
                iteration_n += 1
                await self.__wait()
            
            ticks_transformed = []
            for tck in ticks_arr:
                ticks_transformed.append([int(tck.timestamp.timestamp() * 1000), tck.ask[0].price])
            df = pd.DataFrame(ticks_transformed)
            df = df.iloc[::-1]

        else:
            limit = 5000
            exante_interval = self.__get_exante_interval(interval).value
            floating_start = time_start
            klines_arr = []
            iteration_n = 0
            while True:
                if floating_start >= time_stop: break
                klines = self.client.get_ohlc(symbol=instrument, 
                                    duration=exante_interval, 
                                    start = floating_start, 
                                    end = time_stop, 
                                    limit = limit, 
                                    agg_type=DataType.QUOTES)
                if klines == None:
                    self._raise_error("Error while downloading klines")
                klines_arr = klines + klines_arr
                if klines == []:
                    if iteration_n == 0:
                        self._raise_error("Error while downloading klines ...")
                if len(klines) == 5000:
                    floating_start = int(klines[0].timestamp.timestamp() * 1000) + (exante_interval * 1000)

                if len(klines)<5000:
                    break
                iteration_n += 1
                await self.__wait()
            
            # klines_transformed = []
            # for kl in klines_arr:
            #     # klines_transformed.append([int(round(datetime.timestamp(kl.timestamp) * 1000)), kl.open_])
            #     klines_transformed.append([int(kl.timestamp.timestamp() * 1000), kl.open_])
            # df = pd.DataFrame(klines_transformed)
            # df = df.iloc[::-1]

            klines_transformed = []
            exante_interval = self.__get_exante_interval(interval).value
            prev_kl = None
            for kl in reversed(klines_arr):
                if prev_kl and kl.timestamp.timestamp() * 1000 - prev_kl.timestamp.timestamp() * 1000 >= exante_interval * 1000*2:
                    new_timestamp = int(prev_kl.timestamp.timestamp() * 1000)+ (exante_interval * 1000)
                    new_datetime = datetime.utcfromtimestamp(new_timestamp/1000)
                    klines_transformed.append([new_timestamp, prev_kl.close]) 
                klines_transformed.append([int(kl.timestamp.timestamp() * 1000), kl.open_])
                prev_kl = kl
            df = pd.DataFrame(klines_transformed)
            # self._log('exante shape before validation', df.shape)
            df = validate_dataframe_timestamps(df, self.__get_exante_interval(interval).value * 1000, 
                    time_start, time_stop)
            # self._log('exante shape after validation', df.shape)
            # print('head', str(df.head(1)))
            # print('tail', str(df.tail(1)))


        return df   



import os
from backtesterrb30.libs.utils.module_loaders import import_model_module
# here = os.getcwd()
# print('ssssssssssssssssssssssss',os.environ['STRATEGY_FILE'])
strategy_path = os.getenv('STRATEGY_PATH')
strategy_file = os.getenv('STRATEGY_FILE')
model_class_name = os.getenv('MODEL_CLASS_NAME')
from backtesterrb30.libs.utils.run_service import run_service
microservice_name = 'python_engine'
module = import_model_module(strategy_path, strategy_file, model_class_name)
run_service(microservice_name, module)


from importlib.machinery import SourceFileLoader
from os.path import join
import importlib.util


def import_data_schema(path_to_strategy: str,strategy_file: str, data_class_name: str):
    module = SourceFileLoader('data', path_to_strategy+'/'+strategy_file).load_module()
    data_class = getattr(module, data_class_name)
    return data_class

def import_model_module(path_to_strategy: str,strategy_file: str, model_class_name: str):
    module = SourceFileLoader('model', path_to_strategy+'/'+strategy_file).load_module()
    model_class = getattr(module, model_class_name)
    return model_class

def import_executor_module(path_to_strategy: str, strategy_file: str, executor_class_name: str):
    module = SourceFileLoader('executor', path_to_strategy+'/'+strategy_file).load_module()
    executor_class = getattr(module, executor_class_name)
    return executor_class

def import_spec_module(path_to_module):
    spec=importlib.util.spec_from_file_location(path_to_module, path_to_module)
    module = importlib.util.module_from_spec(spec)
    return spec, module

def reload_spec_module(spec, module):
    spec.loader.exec_module(module)


import unittest
from  historical_data_feeds.historical_data_feeds import HistoricalDataFeeds

class TestConfig:
    name = "test",
    strategy_path = 'hackaton1'

class TestHistoricalDataFeeds(unittest.TestCase):

    @classmethod
    def setUpClass(cls) -> None:
        print('setup class')

    @classmethod
    def tearDownClass(cls) -> None:
        print('teardown class')

    def setUp(self) -> None:

        config = TestConfig()
        self.historical_data_feeds = HistoricalDataFeeds(config)
        

    def tearDown(self) -> None:
        return super().tearDown()

    def test_get_next_value(self):
        result = self.historical_data_feeds.get_next_value()
        self.assertEqual(result, False)



if __name__ == '__main__':
    unittest.main()



import pandas as pd

def validate_dataframe_timestamps(df: pd.DataFrame, 
            interval_step_miliseconds: int, time_start: int, time_stop: int) -> pd.DataFrame:
    try:
        if interval_step_miliseconds > 1000*60*60*24: 
            print('cannot validate df with this interval')
            return df
        timestamp_interval = interval_step_miliseconds
        list_timestamps = range(time_start, time_stop, timestamp_interval)
        for i, timestamp in enumerate(list_timestamps):
            if i == 0: continue
            count = 0
            while int(df.iloc[i][0]) != int(timestamp):
                count+= 1
                df = pd.concat([df.iloc[:i],df.iloc[i-1:i],df.iloc[i:]], ignore_index=True)
                df.iloc[i,0] = int(timestamp)
    except IndexError:
        pass
    return df



from pydantic import BaseModel
from typing import Optional, List
from backtesterrb30.libs.interfaces.python_backtester.custom_chart import CustomChart


class DebugBreakpoint(BaseModel):
    custom_charts: Optional[List[CustomChart]] = None
    display_charts: Optional[bool] = True


from backtesterrb30.libs.interfaces.utils.data_schema import DataSchema
from backtesterrb30.libs.interfaces.utils.data_symbol import DataSymbol
from backtesterrb30.libs.data_sources.data_source_base import DataSource
from backtesterrb30.libs.interfaces.historical_data_feeds.instrument_file import InstrumentFile
from backtesterrb30.historical_data_feeds.functions import create_downloading_clients, check_symbol_data_exists
import asyncio
from typing import List


class HistoricalDownloader():
    def __init__(self, data_schema: DataSchema, historical_sources_array, downloaded_data_path, logger=print):
        self._log = logger
        self.__data_schema: DataSchema = data_schema
        self.__historical_sources_array = historical_sources_array
        self.__data_sources = {}
        self.__data_to_download_2 = None
        self.__downloaded_data_path = downloaded_data_path


    def run(self, loop):
        self.__get_data_to_download(self.__data_schema)
        create_downloading_clients(self.__historical_sources_array, self.__data_schema, self.__data_sources, self._log)
        self.__validate_data_schema_instruments(self.__data_schema.data, loop)
        self.__validate_symbols_to_download(self.__data_schema, loop)
        self.__download_prepared_data(self.__data_schema, loop)
        return self.__data_to_download_2


    def __get_data_to_download(self, data_schema: DataSchema):
        for data_symbol in data_schema.data:
            data_symbol.additional_properties['files_to_download']: List[InstrumentFile] = []
            data_symbol.additional_properties['files_to_download'] = check_symbol_data_exists(data_symbol, self.__downloaded_data_path)
            if self.__data_to_download_2 == None:
                self.__data_to_download_2 = []
            self.__data_to_download_2 = self.__data_to_download_2 + data_symbol.additional_properties['files_to_download']


    def __validate_symbols_to_download(self,data_schema: DataSchema, loop: asyncio.AbstractEventLoop):
        self._log('validating symbols before download')
        for symbol in data_schema.data:
            if symbol.additional_properties['files_to_download'] != []:
                data_source_client: DataSource = self.__get_data_source_client(symbol.historical_data_source)
                res = loop.run_until_complete(data_source_client.validate_instrument(symbol))
                if not res:
                    raise Exception('Error while validation symbol.')

    def __download_prepared_data(self, data_schema: DataSchema, loop: asyncio.AbstractEventLoop):
        for symbol in data_schema.data:
            if symbol.additional_properties['files_to_download'] != []:
                loop.create_task(self.__download_symbol_data(symbol))


    def __validate_data_schema_instruments(self, data_symbol_array: List[DataSymbol], loop: asyncio.AbstractEventLoop):
        self._log('Data_schema validation')
        #check for duplicates:
        seen = []
        for x in data_symbol_array:
            data = str(x.symbol)+str(x.historical_data_source)
            if data in seen:
                raise Exception('Duplicate data symbols')
            else:
                seen.append(data)
        #check other
        number_of_trigger_feeders = 0
        for data in data_symbol_array:
            if data.backtest_date_start == None:
                raise Exception('Error. You must provide "backtest_date_start" field in data_schema file while you are backtesting your strategy')
            if data.backtest_date_start >= data.backtest_date_stop: 
                raise Exception('Error. You have provided "backtest_date_start" is equal or bigger than "backtest_date_start" ')
            if [data.backtest_date_start.hour,
                data.backtest_date_start.minute,
                data.backtest_date_start.second,
                data.backtest_date_start.microsecond] != [0,0,0,0]:
                raise Exception('Error. Provide your "backtest_date_start" and "backtest_date_stop" in a day accuracy like: "backtest_date_start": datetime(2020,6,1)')
            if data.historical_data_source not in self.__historical_sources_array: 
                raise Exception('Error. This historical_data_source not implemented yet')
            if data.trigger_feed == True:
                number_of_trigger_feeders += 1
        if number_of_trigger_feeders < 1:
            raise Exception('Error. Your "data_schema.py" must have at least one instrument that triggers feeds')
    
    def __get_data_source_client(self, historical_source: str) -> DataSource:
        client = self.__data_sources[historical_source]
        if not client:
                raise Exception('Error, no registered source client')
        return client

    async def __download_symbol_data(self, symbol: DataSymbol):
        files_to_download: List[InstrumentFile] = symbol.additional_properties['files_to_download']
        await self.__download_symbol_files(files_to_download, symbol.historical_data_source)

    async def __download_symbol_files(self, files: List[InstrumentFile], source: str):
        data_source_client: DataSource = self.__get_data_source_client(source)
        for file in files:
            await data_source_client.download_instrument(self.__downloaded_data_path, file)





from pydantic import BaseModel, validator
from typing import List, Any
from pandas import DataFrame
from backtesterrb30.libs.interfaces.python_backtester.trade import Trade
from backtesterrb30.libs.interfaces.utils.data_symbol import DataSymbol

class Position(BaseModel):
    number_of_actions: float = 0
    position_outcome: float = 0
    buy_summary_cost: float = 0
    sell_summary_cost: float = 0
    last_instrument_price: float = 0
    trades: List = []



import os
from backtesterrb30.libs.utils.list_of_services import SERVICES_ARRAY
serve_file_commands_array = [
    'from sys import argv\n',
    'from dotenv import load_dotenv\n',
    'load_dotenv(".env")\n',
    'load_dotenv(".additional_configs")\n',
    'microservice_name = argv[1]\n',
    'if microservice_name == "python_backtester":\n',
    '    from backtesterrb30.historical_data_feeds import run\n',
    'elif microservice_name == "python_engine":\n',
    '    from backtesterrb30.python_backtester import run\n',
    'elif microservice_name == "python_executor":\n',
    '    from backtesterrb30.python_engine import run\n',
    'elif microservice_name == "historical_data_feeds":\n',
    '    from backtesterrb30.python_executor import run\n'
    'elif microservice_name == "live_data_feeds":\n',
    '    from backtesterrb30.live_data_feeds import run\n'
]

run_file_commands_array = [
    'sudo -v\n'
    'trap ctrl_c INT\n'
    'function ctrl_c() {\n'
    '        echo "** Trapped CTRL-C"\n'
    '        sudo killall python3 \n'
    '        exit 2\n'
    '}\n'
    'sudo python3 serve.py python_backtester &\n'
    'sudo python3 serve.py python_engine &\n'
    'sudo python3 serve.py python_executor &\n'
    'sudo python3 serve.py historical_data_feeds &\n'
    'sudo python3 serve.py live_data_feeds &\n'
    'while true\n'
    'do\n'
    '    sleep 10\n'
    'done\n'
]

def run_all_microservices(
            strategy_path, 
            backtest_state,
            strategy_file,
            model_class_name,
            executor_class_name,
            data_class_name):
    serve_prepared = False
    run_prepared = False
    for fname in os.listdir('.'):
        if fname == 'serve.py':
            serve_prepared = True
        if fname == 'run.sh':
            run_prepared = True
    if serve_prepared == False:
        with open('serve.py', 'w') as f:
            for line in serve_file_commands_array:
                f.write(line)
    if run_prepared == False:
        with open('run.sh', 'w') as f:
            for line in run_file_commands_array:
                f.write(line)

    print('strategy_file', strategy_file)
    services_array = SERVICES_ARRAY

    def is_port_in_use(port: int) -> bool:
        import socket
        with socket.socket(socket.AF_INET, socket.SOCK_STREAM) as s:
            return s.connect_ex(('localhost', port)) == 0

    def create_port_configurations():
        start_port = 1000
        assigned = False
        while not assigned:
            port_in_use = False
            for port in range(start_port,start_port+50):
                if is_port_in_use(port):
                    print('Port '+str(port)+ ' is already in use, changing port range.')
                    start_port += 50
                    port_in_use = True
                    break
            if port_in_use == False:
                assigned = True
                
        with open('.additional_configs', 'w') as f:
            for i, service in enumerate(services_array):
                f.write(service+'_pubs='+str(start_port+i)+'\n')
                subs_str = service+'_subs='
                for j in range(len(services_array)):
                    if j != i:
                        subs_str += str(start_port+j)+','
                subs_str = subs_str[:-1]
                subs_str += '\n'
                f.write(subs_str)
            
    
    print('preparing microservice ports configuration')
    create_port_configurations()
    bck_state = 'true' if backtest_state else 'false'
    with open('.additional_configs', 'a') as f:
        f.write('backtest_state='+bck_state+'\n')
        f.write('STRATEGY_FILE='+strategy_file+'\n')
        f.write('STRATEGY_PATH='+strategy_path+'\n')
        f.write('MODEL_CLASS_NAME='+model_class_name+'\n')
        f.write('EXECUTOR_CLASS_NAME='+executor_class_name+'\n')
        f.write('DATA_CLASS_NAME='+data_class_name+'\n')

    if backtest_state:
        os.system('bash run.sh') 
    else:
        print('live strategies not implemented')
        
    # print('starting all containers')
    # system('docker-compose up --build --remove-orphans')



from typing import Union
from backtesterrb30.libs.data_sources.data_source_base import DataSource
from datetime import datetime
from backtesterrb30.libs.interfaces.historical_data_feeds.instrument_file import InstrumentFile
from os import getenv
from backtesterrb30.libs.interfaces.utils.data_symbol import DataSymbol
from backtesterrb30.libs.utils.singleton import singleton
# from backtesterrb30.historical_data_feeds.modules.utils import validate_dataframe_timestamps
import pandas as pd

class RB30DataSource(DataSource):
    INTERVALS= {}
    NAME='rb30disk'
    
    def __init__(self, logger=print):
        super().__init__(False, logger)

    def _get_interval_miliseconds(self, interval: str) -> Union[int,None]: 
        pass

    async def _validate_instrument_data(self, data: DataSymbol) -> bool:
        return True

    async def _download_instrument_data(self, 
                    instrument: str, interval: str, time_start: int, time_stop: Union[int, None]) -> pd.DataFrame:
        return True



from abc import abstractmethod
from typing import Union
from backtesterrb30.libs.interfaces.historical_data_feeds.instrument_file import InstrumentFile
import asyncio
import pandas as pd
from enum import Enum
from backtesterrb30.libs.interfaces.utils.data_symbol import DataSymbol
from os.path import join

class DataSource():    
    def __init__(self, allow_synchronous_download: bool, logger=print):
        self._allow_synchronous_download = allow_synchronous_download
        self._log = logger

    def _raise_error(self, error_message):
        raise Exception(error_message)

    def __validate_dataframe(self, 
                        df: pd.DataFrame) -> bool:
        # if int(df.head(1)[0]) != expected_first_data_timestamp: return False
        # if int(df.tail(1)[0]) != expected_last_data_timestamp: return False
        # if len(df) != expected_data_length: return False
        if df.shape[1] != 2:
            raise Exception('Bad shape of returned dataframe in data source '+self.NAME)
        return True

    async def download_instrument(self,
                        downloaded_data_path: str, 
                        instrument_file: InstrumentFile): 
        await asyncio.sleep(0.1)
        df: pd.DataFrame = await self._download_instrument_data(instrument_file.instrument,
                instrument_file.interval, instrument_file.time_start, instrument_file.time_stop)
        self.__validate_dataframe(df)
        df.to_csv(join(downloaded_data_path, instrument_file.to_filename()), 
        index=False, header=False)
        
    async def download_dataframe(self, instrument: str, interval: str, 
            time_start: int, time_stop: Union[int, None]) ->pd.DataFrame:
        await asyncio.sleep(0.1)
        df: pd.DataFrame = await self._download_instrument_data(instrument,
                interval, time_start, time_stop)
        self.__validate_dataframe(df)
        return df


    async def validate_instrument(self, data: DataSymbol) -> bool:
        await asyncio.sleep(0.1)
        if type(data.interval) != self.INTERVALS:
            raise Exception('Bad interval type in data symbol', data.symbol, data.historical_data_source, 'check data_schema.py')
        return await self._validate_instrument_data(data)

    @abstractmethod
    def _get_interval_miliseconds(self, interval: str) -> Union[int,None]: 
        """
        This function should return number of miliseconds delay for provided interval
        """
        pass

    @abstractmethod
    async def _validate_instrument_data(self, data: DataSymbol) -> bool:
        """
        Function should validate if data for shis DataSymbol is possible to download.
        It should check if
        - data.symbol exists
        - data.interval fot this data.symbol exists
        - timestamps range from 'data.backtest_date_start' to 'data.backtest_date_stop' 
            for this data.symbol and data.interval exists
        """
        pass
    
    @abstractmethod
    async def _download_instrument_data(self,
                instrument: str, interval: str, time_start: int, time_stop: Union[int, None]) -> pd.DataFrame:
        """
        First downloaded timestamp should exactly exuals instrument_file.time_start (can be bigger if u are downloading ticks)
        Last downloaded timestamp should exactly exuals one interval before instrument_file.time_stop ()
        """
        pass


    def get_current_price(self, data: DataSymbol):
        """
        Function should return current price of provided data symbol
        """
        pass

    


from pydantic import BaseModel
from typing import Optional, List
from backtesterrb30.libs.interfaces.python_backtester.custom_chart import CustomChart


class DataFinish(BaseModel):
    # main_instrument_price: float
    custom_charts: Optional[List[CustomChart]] = None




from pydantic import BaseModel
from typing import List
from pandas import DataFrame

class ExecutorPosition(BaseModel):
    instrument: str
    number_of_actions: float = 0
    position_size_in_usd: float = 0
    last_instrument_price: float = 0




from abc import abstractmethod
import asyncio
from backtesterrb30.libs.communication_broker.broker_base import BrokerBase
from backtesterrb30.libs.interfaces.python_backtester.data_start import DataStart
from backtesterrb30.libs.interfaces.python_backtester.debug_breakpoint import DebugBreakpoint
from backtesterrb30.libs.interfaces.python_backtester.last_feed import LastFeed
from backtesterrb30.libs.interfaces.python_backtester.positions import Position
from backtesterrb30.libs.interfaces.utils.data_symbol import DataSymbol
from backtesterrb30.libs.communication_broker.zmq_broker import ZMQ
from backtesterrb30.libs.communication_broker.asyncio_broker import AsyncioBroker
from backtesterrb30.libs.interfaces.python_backtester.trade import Trade
from backtesterrb30.libs.interfaces.python_backtester.money_state import MoneyState
from backtesterrb30.libs.interfaces.python_backtester.data_finish import DataFinish, CustomChart
from backtesterrb30.libs.interfaces.utils.data_schema import DataSchema
from backtesterrb30.libs.utils.list_of_services import SERVICES, SERVICES_ARRAY
import pandas as pd
from os.path import join
import matplotlib.pyplot as plt
from matplotlib.ticker import ScalarFormatter
import time as tm
import numpy as np
from typing import List, Union
from datetime import datetime
from backtesterrb30.libs.utils.service import Service
from backtesterrb30.libs.interfaces.utils.config import Config, BROKERS
from appdirs import user_cache_dir

class Backtester(Service):
    
    downloaded_data_path = user_cache_dir('rb30_cache')
    _broker: BrokerBase

    def __init__(self, config: Config, data_schema: DataSchema, loop = None, logger=print):
        super().__init__(config, logger)
        self.config: Config=config
        self.__loop =  loop
        self.__custom_event_loop = False
        if self.__loop == None: 
            self.__loop = asyncio.get_event_loop()
            self.__custom_event_loop= True
        self.data_schema: DataSchema = data_schema
        for dat in self.data_schema.data:
            dat.additional_properties['position'] = None
            dat.additional_properties['chart_data_frame'] = None
        
        # self.trading_instruments_charts = []
        self.cumulated_money_chart = []
        # self.positions: List[Position] = []
        self.__last_feed: LastFeed = {}
        self.__last_timestamp: int = 0

        self.__backtest_start_time = 0
        self.__fig = None
        self.__chart_displayed = False



    # # override
    # def _asyncio_loop(self, loop: asyncio.AbstractEventLoop):
    #     self._create_listeners(loop)
    #     loop.create_task(self.__update_chart())

    # # override
    # def _handle_zmq_message(self, message):
    #     pass

    def _loop(self):
        # self._broker.run()
        self._broker.create_listeners(self.__loop)
        self.__loop.create_task(self.__update_chart())
        if self.__custom_event_loop:
            self.__loop.run_forever()
            self.__loop.close()

    # def _send(self, service: SERVICES, msg: str, *args):
    #     self._broker.send(service, msg, *args)

    def _configure(self):
        super()._configure()
        self._broker.register("trade", self.__trade_event)
        self._broker.register("data_finish", self.__data_finish_event)
        self._broker.register("close_all_trades", self.__close_all_trades_event)
        self._broker.register("data_start", self.__data_start_event)
        self._broker.register("debug_breakpoint", self.__debug_breakpoint_event)
        self._broker.register("last_feed", self.__last_feed_event)

    # def _trigger_event(self, event):
    #     pass
    
    async def __stop_all_services(self):
        for service in SERVICES_ARRAY:
            if service != self.name:
                await self._broker.send(getattr(SERVICES, service), 'stop')
        self._broker.stop()

    def __axis_format(self, ax, title):
        ax.set_title(title)
        ax.set_axisbelow(True)
        ax.yaxis.grid(color='gray', linestyle='dashed')


    async def __print_charts(self,
                    custom_charts: List[CustomChart] ):
        plt.close()
        
        #plot trades chart
        # position: Position = self.positions[0]
        number_of_custom_charts = 0
        if custom_charts != None:
            number_of_custom_charts = len([ch for ch in custom_charts if not ch.display_on_price_chart])

        chartable_symbols = [sym for sym in self.data_schema.data \
                if sym.display_chart_in_summary == True]
        axs_number_used = 0

        #prepare axes
        self.__fig, axs = plt.subplots(nrows=1+ len(chartable_symbols)+number_of_custom_charts, \
                ncols=1, sharex = True, figsize=(13, 13))
        if type(axs)!= np.ndarray:
            # if only 1 axis its returned not in array
            axs = [axs]
        ax = None
        # plot instrment charts
        for sym in chartable_symbols:
            position: Position = sym.additional_properties['position']
            # if position:
            if self.__last_timestamp != None:
                main_chart = sym.additional_properties['chart_data_frame'].loc[sym.additional_properties['chart_data_frame']['timestamp'] <= \
                        self.__last_timestamp]
            ax = main_chart.plot(x ='timestamp', y='price', kind = 'line', ax=axs[axs_number_used])
            axs_number_used += 1
            self.__axis_format(ax, sym.symbol +" - "+sym.historical_data_source)
            if self.data_schema.log_scale_valuation_chart:	
                ax.yaxis.set_major_formatter(ScalarFormatter())
            if position:
                normalized_quants = self.__normalize([abs(trade[2]) for trade in position.trades], (5,15))   
                for trade, quant in zip(position.trades, normalized_quants):
                    ax.plot(trade[0], trade[1], '.g' if trade[2]>0 else '.r', ms=quant)

        if len(self.cumulated_money_chart) > 0:
            #plot money chart
            money_df = pd.DataFrame(self.cumulated_money_chart, columns=['timestamp', 'income'])
            ax = money_df.plot(x ='timestamp', y='income', kind = 'line', ax=axs[axs_number_used], sharex = ax)
            axs_number_used += 1
            
        #plot custom charts
        if custom_charts != None:
            for i, ch in enumerate(custom_charts):
                chart = [[c.timestamp, c.value] for c in ch.chart ]
                custom_df = pd.DataFrame(chart, columns=['timestamp', ch.name])
                if ch.display_on_price_chart:
                    custom_df.plot(x ='timestamp', y=ch.name, kind = 'line', ax=axs[0], \
                            sharex = ax, color = ch.color)
                else:
                    ax = custom_df.plot(x ='timestamp', y=ch.name, kind = 'line', ax=axs[axs_number_used], \
                            sharex = ax, color = ch.color)
                    axs_number_used += 1
                    self.__axis_format(ax, ch.name)
                    if ch.log_scale:
                        ax.yaxis.set_major_formatter(ScalarFormatter())
        if self.__chart_displayed == False:
            plt.ion()
            plt.show(block = False)
            self.__chart_displayed = True


    def __add_position(self, data_symbol: DataSymbol) -> Position:
        pos = Position()
        data_symbol.additional_properties['position'] = pos
        self.__update_last_instrument_prices()
        return pos


    async def __update_chart(self): 
        while True:
            if self.__fig != None:
                try:
                    self.__fig.canvas.draw()
                    self.__fig.canvas.flush_events()
                except Exception as e: 
                    pass
            await asyncio.sleep(0.1)

    def __normalize(self, x, newRange=(0, 1)): #x is an array. Default range is between zero and one
        if len(x) == 0:
            return []
        xmin, xmax = np.min(x), np.max(x) #get max and min from input array
        if xmin != xmax:
            norm = (x - xmin)/(xmax - xmin) # scale between zero and one

            if newRange == (0, 1):
                return(norm) # wanted range is the same as norm
            elif newRange != (0, 1):
                return norm * (newRange[1] - newRange[0]) + newRange[0] #scale to a different range.
            #add other conditions here. For example, an error messag
        return np.ones_like(x) * 15

    async def __recalculate_positions(self):
        positions: List[Position] = [elem.additional_properties['position'] for elem in self.data_schema.data]
        positions = [pos for pos in positions if pos != None]
        for pos in positions:
            pos.position_outcome =  - pos.buy_summary_cost - pos.sell_summary_cost + \
                    pos.number_of_actions * pos.last_instrument_price
        current_capital = sum([pos.position_outcome for pos in positions])
        current_invested = sum([abs(pos.number_of_actions) * pos.last_instrument_price\
                for pos in positions])
        self.cumulated_money_chart.append([self.__last_timestamp, current_capital])
        await self._broker.send(SERVICES.python_executor, 'set_current_capital_event', current_capital)
        await self._broker.send(SERVICES.python_executor, 'set_current_invested_event', current_invested)

    def __get_data_symbol(self, symbol: str, source: str):
        arr = [sym for sym in self.data_schema.data if sym.symbol ==\
                symbol and sym.historical_data_source == source]
        if len(arr) == 0:
            raise Exception('No data symbol found')
        if len(arr) > 1: 
            raise Exception('More than one symbol found')
        return arr[0]

    async def __trade(self, trade: Trade):
        symbol: DataSymbol = self.__get_data_symbol(trade.symbol, trade.source)
        position = symbol.additional_properties['position']
        if not position:
            position = self.__add_position(symbol)
        if trade.price == 0 and trade.timestamp == 0:
            trade.price = position.last_instrument_price
            trade.timestamp = self.__last_timestamp
        if trade.price <= 0 or trade.timestamp <= 0:
            self._log('Cannot make trade with this price and timestamp:', trade.price,';', trade.timestamp)
            return
        self._log(f"Trade for [{symbol.symbol}]: time={datetime.utcfromtimestamp(trade.timestamp/1000)}, | value={trade.value}, price={trade.price}")
        position.trades.append([trade.timestamp, trade.price, trade.value])
        # print('trade price', trade.price)
        position.number_of_actions += trade.value / trade.price
        if trade.value > 0: 
            position.buy_summary_cost += trade.value
        else:
            position.sell_summary_cost += trade.value
        await self.__recalculate_positions()
        

    async def __print_summary(self, 
                custom_charts: List[CustomChart], 
                display_charts = True
                # file_names: List[str],
                # main_instrument_price: float,
                # last_timestamp: Union[int, None] = None
                ):
        self._log('==========================')
        self._log(('BREAKPOINT ' if self.__last_timestamp != None else '') + 'SUMMARY')
        finish_time = tm.time()
        time_of_backtest = finish_time - self.__backtest_start_time
        self._log('time of backtest:', round(time_of_backtest,2), '[s]')
        # self._log('number of trades:', len(self.trades))
        # self._log('buy_summary_cost:', self.buy_summary_cost)
        # self._log('sell_summary_cost:', self.sell_summary_cost)
        # self._log('number of unrealized actions:', self.number_of_actions)
        # self._log('biggest investment: ', self.biggest_investment)
        # self._log('actual price:', main_instrument_price)
        # income = - self.buy_summary_cost - self.sell_summary_cost + self.number_of_actions * main_instrument_price
        if len(self.cumulated_money_chart) > 0:
            self._log('income:', self.cumulated_money_chart[-1][1])
        self._log('==========================')
        self._log('')
        if display_charts:
            await self.__print_charts(custom_charts)

    def __load_instrument_chart(instrument):
        loaded_df = None
        
        return loaded_df

    # COMMANDS

    async def __data_finish_event(self, finish_params):
        finish_params = DataFinish(**finish_params)
        self._log('')
        self._log('====================================================')
        self._log('BACKTEST FINISHED')
        await self.__print_summary(finish_params.custom_charts)
        self._log('====================================================')
        plt.show(block = True)
        await self.__stop_all_services()

    async def __trade_event(self, msg):
        trade: Trade = Trade(**msg)
        await self.__trade(trade)

    async def __last_feed_event(self, msg):
        # print('last feed event')
        self.__last_feed = LastFeed(**msg)
        self.__last_timestamp = self.__last_feed.last_feed[0]
        self.__update_last_instrument_prices()

    def __update_last_instrument_prices(self):
        for data_symbol, last_feed in zip(self.data_schema.data, self.__last_feed.last_feed[1:]):
            # print('updating pos')

            position: Position = data_symbol.additional_properties['position']
            if position:
                # print('upda')
                # print(type(last_feed))
                if type(last_feed) == float or type(last_feed) == int: 
                    position.last_instrument_price = last_feed
                    continue 
                if type(last_feed) == list:
                    print('feed is 3d')
                    if type(last_feed[0]) == float or type(last_feed[0]) == int:
                        position.last_instrument_price = last_feed[0]
                    continue

                raise Exception('Error during setting last instrument price')

    async def __close_all_trades_event(self):
        self._log(f"Received close all trades command NOT IMPLEMENTED")

        #TODO

    async def __data_start_event(self, start_params):
        start_params = DataStart(**start_params)
        self.__backtest_start_time = start_params.start_time
        
        for element in start_params.file_names:
            dfs = []
            for file in element.files:
                df = pd.read_csv(join(self.downloaded_data_path, file.to_filename()), \
                        index_col=None, header=None, names=['timestamp', 'price'])
                dfs.append(df)
            symbol = self.__get_data_symbol(element.symbol, element.source)
            symbol.additional_properties['chart_data_frame'] = pd.concat(dfs)


    async def __debug_breakpoint_event(self, breakpoint_params):
        breakpoint_params = DebugBreakpoint(**breakpoint_params)
        await self.__print_summary(breakpoint_params.custom_charts, breakpoint_params.display_charts)





from backtesterrb30.libs.utils.run_all import run_all_microservices
from backtesterrb30.libs.interfaces.python_engine.custom_chart_element import CustomChartElement
from backtesterrb30.python_engine.engine import Engine
from backtesterrb30.python_executor.trade_executor import Executor
from backtesterrb30.libs.utils.config_validator import validate_config
from backtesterrb30.libs.data_sources.data_sources_list import HISTORICAL_SOURCES
from backtesterrb30.libs.utils.strategy import Strategy
from backtesterrb30.libs.interfaces.utils.data_schema import DataSchema
# from backtesterrb30.libs.utils.historical_sources import(
#     EXANTE_INTERVALS, 
#     HISTORICAL_SOURCES, 
#     BINANCE_INTERVALS, 
#     DUKASCOPY_INTERVALS, 
#     COINGECKO_INTERVALS)
# from backtesterrb30.libs.utils.data_imports import *
# from backtesterrb30.libs.utils.model_imports import *
# from backtesterrb30.libs.utils.executor_imports import *


from backtesterrb30.libs.utils.run_service import run_service
microservice_name = 'historical_data_feeds'
from backtesterrb30.historical_data_feeds.historical_data_feeds import HistoricalDataFeeds
run_service(microservice_name, HistoricalDataFeeds)